# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a
LightGBM gradient boosting regressor
into a :gurobipy:`model`.
"""


import numpy as np
from gurobipy import GRB

from ..exceptions import NoSolution
from ..modeling import AbstractPredictorConstr
from ..modeling.decision_tree import AbstractTreeEstimator


def add_lgbmregressor_constr(
    gp_model, lgbm_regressor, input_vars, output_vars=None, epsilon=0.0, **kwargs
):
    """Formulate lgbm_regressor into gp_model.

    The formulation predicts the values of output_vars using input_vars
    according to lgbm_regressor. See our :ref:`User's Guide
    <Gradient Boosting Regression>` for details on the mip formulation used.

    This version is for using directly with the Scikit-Learn wrapper of lgbm.
    Note that only "gbtree" regressors are supported at this point.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    lgbm_regressor : :external+lightgbm:py:class:`lightgbm.sklearn.LGBMRegressor`
        The gradient boosting regressor to insert as predictor.
    input_vars : mvar_array_like
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : mvar_array_like, optional
        Decision variables used as output for gradient boosting regressor in model.

    Returns
    -------
    LightGBMRegressorConstr
        Object containing information about what was added to gp_model to formulate
        gradient_boosting_regressor.

    Notes
    -----
    |VariablesDimensionsWarn|

    Also see
    :py:func:`gurobi_ml.sklearn.decision_tree_regressor.add_decision_tree_regressor`
    for specific parameters to model decision tree estimators.

    Raises
    ------
    NoModel
        If the booster is not of type "gbtree".
    """
    return LGBMConstr(
        gp_model,
        lgbm_regressor.booster_,
        input_vars,
        output_vars,
        epsilon=epsilon,
        **kwargs,
    )


def add_lgbm_booster_constr(
    gp_model, lgbm_booster, input_vars, output_vars=None, epsilon=0.0, **kwargs
):
    """Formulate lgbm_booster into gp_model.

    The formulation predicts the values of output_vars using input_vars
    according to lgbm_regressor. See our :ref:`User's Guide
    <Gradient Boosting Regression>` for details on the mip formulation used.

    Note that only "gbtree" regressors are supported at this point.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    lgbm_regressor : :external+lightgbm:py:class:`lightgbm.Booster`
        The booster to insert as predictor.
    input_vars : mvar_array_like
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : mvar_array_like, optional
        Decision variables used as output for gradient boosting regressor in model.

    Returns
    -------
    LightGBMRegressorConstr
        Object containing information about what was added to gp_model to formulate
        gradient_boosting_regressor.

    Notes
    -----
    |VariablesDimensionsWarn|

    Also see
    :py:func:`gurobi_ml.sklearn.decision_tree_regressor.add_decision_tree_regressor`
    for specific parameters to model decision tree estimators.

    Raises
    ------
    NoModel
        If the booster is not of type "gbtree".
    """
    return LGBMConstr(
        gp_model, lgbm_booster, input_vars, output_vars, epsilon=epsilon, **kwargs
    )


class LGBMConstr(AbstractPredictorConstr):
    """Class to model trained :external+lightgbm:py:class:`lightgbm.Booster`
    in a gurobipy model.

    |ClassShort|
    """

    def __init__(
        self, gp_model, lgbm_regressor, input_vars, output_vars, epsilon=0.0, **kwargs
    ):
        self._output_shape = 1
        self.estimators_ = []
        self.lgbm_regressor = lgbm_regressor
        self._default_name = "lgbm_reg"
        self.epsilon = epsilon
        AbstractPredictorConstr.__init__(
            self, gp_model, input_vars, output_vars, **kwargs
        )

    @staticmethod
    def _count_nodes(root_node):
        """Count the nodes in a lightgbm tree.

        Traverse the tree and count the number of leaf and split nodes.

        Returns
        -------
        num_leafs : int
            Number of leaf nodes in the tree.
        num_split : int
            Number of split nodes in the tree.
        """
        heap = [root_node]
        num_leafs = 0
        num_split = 0
        while len(heap) > 0:
            node = heap.pop()
            if "split_index" in node:
                heap.append(node["left_child"])
                heap.append(node["right_child"])
                num_split += 1
            else:
                num_leafs += 1
        return (num_leafs, num_split)

    @staticmethod
    def _flat_tree_representation(root_node):
        """Flatten a lightgbm tree.

        This function takes a root node of a lightgbm tree and flattens it into a dictionary representation.
        The flattened tree contains information about the children nodes, split features, thresholds, and leaf values.

        Args:
            root_node (dict): The root node of the lightgbm tree.

        Returns:
            dict: A dictionary representation of the flattened tree.

        """
        num_leafs, num_split = LGBMConstr._count_nodes(root_node)

        numnodes = num_leafs + num_split
        children_left = np.full(numnodes, -2, dtype=int)
        children_right = np.full(numnodes, -2, dtype=int)
        feature = np.empty(numnodes, dtype=int)
        value = np.empty(numnodes, dtype=np.float32)
        threshold = np.empty(numnodes, dtype=np.float32)

        def node_index(node):
            if "split_index" in node.keys():
                return node["split_index"]
            return node["leaf_index"] + num_split

        heap = [root_node]
        while len(heap) > 0:
            node = heap.pop()
            index = node_index(node)
            assert children_left[index] == -2
            assert children_right[index] == -2
            if "split_index" in node.keys():
                feature[index] = node["split_feature"]
                threshold[index] = node["threshold"]
                children_left[index] = node_index(node["left_child"])
                children_right[index] = node_index(node["right_child"])

                heap.append(node["left_child"])
                heap.append(node["right_child"])
            else:
                children_left[index] = -1
                children_right[index] = -1
                value[index] = node["leaf_value"]

        assert min(children_left) >= -1
        assert min(children_right) >= -1
        flat_tree = {
            "children_left": children_left,
            "children_right": children_right,
            "feature": feature,
            "threshold": threshold,
            "value": value.reshape(-1, 1),
            "capacity": numnodes,
        }
        return flat_tree

    def _mip_model(self, **kwargs):
        """Predict output variables y from input variables X using the
        decision tree.

        Both X and y should be array or list of variables of conforming dimensions.
        """
        model = self.gp_model
        lgbm_regressor = self.lgbm_regressor

        _input = self._input
        output = self._output
        nex = _input.shape[0]
        timer = AbstractPredictorConstr._ModelingTimer()
        outdim = output.shape[1]
        assert (
            outdim == 1
        ), "Output dimension of gradient boosting regressor should be 1"

        lgbm_raw = lgbm_regressor.dump_model()

        trees = lgbm_raw["tree_info"]
        n_estimators = len(trees)

        estimators = []
        if self._no_debug:
            kwargs["no_record"] = True

        tree_vars = model.addMVar(
            (nex, n_estimators, 1),
            lb=-GRB.INFINITY,
            name=self._name_var("esimator"),
        )

        for i, tree in enumerate(trees):
            if self.verbose:
                self._timer.timing(f"Estimator {i}")
            flat_tree = self._flat_tree_representation(tree["tree_structure"])
            flat_tree["n_features"] = lgbm_raw["max_feature_idx"] + 1

            def _name_tree_var(name):
                rval = self._name_var(name)
                if rval is None:
                    return None
                return rval + f"_{i}"

            estimators.append(
                AbstractTreeEstimator(
                    self.gp_model,
                    flat_tree,
                    self.input,
                    tree_vars[:, i, :],
                    self.epsilon,
                    timer,
                    **kwargs,
                )
            )

        self.estimators_ = estimators

        model.addConstr(output == tree_vars.sum(axis=1))

    def print_stats(self, abbrev=False, file=None):
        """Print statistics on model additions stored by this class.

        This function prints detailed statistics on the variables
        and constraints that where added to the model.

        Includes a summary of the estimators that it contains.

        Parameters
        ----------

        file: None, optional
            Text stream to which output should be redirected. By default sys.stdout.
        """
        super().print_stats(abbrev=abbrev, file=file)
        if abbrev or self._no_debug:
            return
        print(file=file)

        # self._print_container_steps("Estimator", self.estimators_, file=file)

    def get_error(self, eps=None):
        if self._has_solution:
            lgbm_in = self.input_values
            lgbm_out = self.lgbm_regressor.predict(lgbm_in)
            r_val = np.abs(lgbm_out.reshape(-1, 1) - self.output.X)
            if eps is not None and np.max(r_val) > eps:
                print(f"{self.output.X} != {lgbm_out.reshape(-1, 1)}")
            return r_val
        raise NoSolution()



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a
:external+sklearn:py:class:`sklearn.linear_model.LogisticRegression` in a
:gurobipy:`model`.
"""

import math
import warnings

import gurobipy as gp
import numpy as np

from ..exceptions import NoModel, ParameterError
from .base_regressions import BaseSKlearnRegressionConstr


def add_logistic_regression_constr(
    gp_model,
    logistic_regression,
    input_vars,
    output_vars=None,
    output_type="classification",
    epsilon=0.0,
    pwl_attributes=None,
    **kwargs,
):
    """Formulate logistic_regression in gp_model.

    The formulation predicts the values of output_vars using input_vars according to
    logistic_regression.

    When the desired output type is classification a mixed integer linear formulation
    using indicator constraints is used. When the desired output type is regression
    then the (non-linear) logistic function is required. For users of Gurobi ≥ 11.0,
    the attribute FuncNonlinear is set to 1 to deal directly with the logistic
    function in an algorithmic fashion.

    For older versions, Gurobi makes a piecewise linear approximation of the logistic
    function.
    The quality of the approximation can be controlled with the parameter
    pwl_attributes. By default, it is parametrized so that the maximal error of the
    approximation is `1e-2`.

    See our :ref:`Users Guide <Logistic Regression>` for
    details on the mip formulation used.

    Parameters
    ----------

    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    logistic_regression : :external+sklearn:py:class:`sklearn.linear_model.LogisticRegression`
        The logistic regression to insert.
    input_vars : mvar_array_like
        Decision variables used as input for logistic regression in model.
    output_vars : mvar_array_like, optional
        Decision variables used as output for logistic regression in model.

    output_type : {'classification', 'probability_1'}, default='classification'
        If the option chosen is 'classification' the output is the class label
        of either 0 or 1 given by the logistic regression. If the option
        'probability_1' is chosen the output is the probability of the class 1.

    epsilon : float, default=0.0
        When the `output_type` is 'classification', this tolerance can be set
        to enforce that class 1 is chosen when the result of the logistic
        function is greater or equal to *0.5 + epsilon*.

        By default, with the value of *0.0*, if the result of the logistic
        function is very close to *0.5* (up to Gurobi tolerances) in the
        solution of the optimization model, the output of the regression can be
        either 0 or 1. The optimization model doesn't make a distinction
        between the two values.

        Setting *esilon* to a small value will remove this ambiguity on the
        output but may also make the model infeasible if the problem is very
        constrained: the open interval *(0.5, 0.5 + epsilon)* is excluded from
        the feasible set of the optimization problem.

    pwl_attributes : dict, optional
        Dictionary for non-default attributes for Gurobi to build the piecewise
        linear approximation of the logistic function. This is only relevent when
        the output type is regression, not classification. The default values for
        those attributes set in the package can be obtained with
        LogisticRegressionConstr.default_pwl_attributes(). The dictionary keys
        should be the `attributes for modeling piece wise linear functions
        <https://www.gurobi.com/documentation/9.1/refman/general_constraint_attribu.html>`_
        and the values the corresponding value the users wants to pass to
        Gurobi.

    Returns
    -------
    LogisticRegressionConstr
        Object containing information about what was added to gp_model to formulate
        logistic_regression.

    Raises
    ------

    NoModel
        If the logistic regression is not a binary label regression

    ParameterError
        If the value of output_type is set to a non-conforming value (see above).

    Notes
    -----
    |VariablesDimensionsWarn|
    """
    return LogisticRegressionConstr(
        gp_model,
        logistic_regression,
        input_vars,
        output_vars,
        output_type,
        epsilon,
        pwl_attributes=pwl_attributes,
        **kwargs,
    )


class LogisticRegressionConstr(BaseSKlearnRegressionConstr):
    """Class to formulate a trained
    :external+sklearn:py:class:`sklearn.linear_model.LogisticRegression` in a gurobipy model.

    |ClassShort|
    """

    def __init__(
        self,
        gp_model,
        predictor,
        input_vars,
        output_vars=None,
        output_type="classification",
        epsilon=0.0,
        pwl_attributes=None,
        **kwargs,
    ):
        if len(predictor.classes_) > 2:
            raise NoModel(
                predictor, "Logistic regression only supported for two classes"
            )
        if output_type not in ("classification", "probability_1"):
            raise ParameterError(
                "output_type should be either 'classification' or 'probability_1'"
            )
        if output_type == "classification" and pwl_attributes is not None:
            message = """
pwl_attributes are not required for classification.  The problem is
formulated without requiring the non-linear logistic function."""
            warnings.warn(message)
        elif output_type != "classification":
            self.attributes = (
                self.default_pwl_attributes()
                if pwl_attributes is None
                else pwl_attributes
            )

        self.epsilon = epsilon
        self._default_name = "log_reg"
        self.affinevars = None
        BaseSKlearnRegressionConstr.__init__(
            self,
            gp_model,
            predictor,
            input_vars,
            output_vars,
            output_type,
            **kwargs,
        )

    @staticmethod
    def default_pwl_attributes() -> dict:
        """Default attributes for approximating the logistic function with Gurobi.

        See `Gurobi's User Manual
        <https://www.gurobi.com/documentation/current/refman/general_constraint_attribu.html>`_
        for the meaning of the attributes.
        """
        if gp.gurobi.version()[0] < 11:
            message = """
Gurobi ≥ 11 can deal directly with nonlinear functions with 'FuncNonlinear'.
Upgrading to version 11 is recommended when using logistic regressions."""
            warnings.warn(message)
            return {
                "FuncPieces": -1,
                "FuncPieceLength": 0.01,
                "FuncPieceError": 0.01,
                "FuncPieceRatio": -1.0,
            }
        return {"FuncNonlinear": 1}

    def _addGenConstrIndicatorMvarV10(self, binvar, binval, lhs, sense, rhs, name):
        """This function is to work around the lack of MVar compatibility in
        Gurobi v10 indicator constraints.  Note, it is not as flexible as Model.addGenConstrIndicator
        in V11+.  If support for v10 is dropped this function can be removed.

        Parameters
        ----------
        binvar : MVar
        binval : {0,1}
        lhs : MVar or MLinExpr
        sense : (char)
            Options are gp.GRB.LESS_EQUAL, gp.GRB.EQUAL, or gp.GRB.GREATER_EQUAL
        rhs : scalar
        name : string
        """
        assert binvar.shape == lhs.shape
        total_constraints = np.prod(binvar.shape)
        binvar = binvar.reshape(total_constraints).tolist()
        lhs = lhs.reshape(total_constraints).tolist()
        for index in range(total_constraints):
            self.gp_model.addGenConstrIndicator(
                binvar[index],
                binval,
                lhs[index],
                sense,
                rhs,
                name=self._indexed_name(index, name),
            )

    def _mip_model(self, **kwargs):
        """Add the prediction constraints to Gurobi."""
        affinevars = self.gp_model.addMVar(
            self.output.shape, lb=-gp.GRB.INFINITY, name="affine_trans"
        )
        self._add_regression_constr(output=affinevars)

        if self.output_type == "classification":
            # For classification we need an extra binary variable
            bin_output = self.gp_model.addMVar(
                self.output.shape, vtype=gp.GRB.BINARY, name="bin_output"
            )

            # Workaround for MVars in indicator constraints for v10.
            addGenConstrIndicator = (
                self.gp_model.addGenConstrIndicator
                if gp.gurobi.version()[0] >= 11
                else self._addGenConstrIndicatorMvarV10
            )

            # The original epsilon is with respect to the range of the logistic function.
            # We must translate this to the domain of the logistic function.
            affine_trans_epsilon = -math.log(1 / (0.5 + self.epsilon) - 1)

            # For classification it is enough to test result of affine transformation
            # and avoid adding logistic curve constraint.  See GH316.
            addGenConstrIndicator(
                bin_output,
                1,
                affinevars,
                gp.GRB.GREATER_EQUAL,
                affine_trans_epsilon,
                "indicator_affinevars_pos",
            )
            addGenConstrIndicator(
                bin_output,
                0,
                affinevars,
                gp.GRB.LESS_EQUAL,
                0,
                "indicator_affinevars_neg",
            )
            self.gp_model.addConstr(bin_output == self.output)
        else:
            log_result = self.output
            for index in np.ndindex(self.output.shape):
                self.gp_model.addGenConstrLogistic(
                    affinevars[index],
                    log_result[index],
                    name=self._indexed_name(index, "logistic"),
                )
            num_gc = self.gp_model.NumGenConstrs
            self.gp_model.update()
            for gen_constr in self.gp_model.getGenConstrs()[num_gc:]:
                for attr, val in self.attributes.items():
                    gen_constr.setAttr(attr, val)
        self.gp_model.update()

    @property
    def affine_transformation_variables(self) -> gp.MVar:
        """Variables that store the result of the affine transformation from the regression coefficient.
        (intermediate result before applying the logistic function).
        """
        return self.affinevars



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Bases classes for modeling neural network layers."""


from .._var_utils import _default_name
from ..base_predictor_constr import AbstractPredictorConstr
from .activations import Identity, ReLU
from .layers import ActivationLayer, DenseLayer


class BaseNNConstr(AbstractPredictorConstr):
    """Base class for inserting a regressor based on a neural-network/tensor into Gurobi.

    This only supports sequential neural networks.

    The bracket operator can be used to iterate over the layers of the layers of the network.
    It will give access to the modeling object of the corresponding layer.

    """

    def __init__(self, gp_model, predictor, input_vars, output_vars, **kwargs):
        self.predictor = predictor
        self.act_dict = {
            "relu": ReLU(),
            "identity": Identity(),
        }
        try:
            for activation, activation_model in kwargs["activation_models"].items():
                self.act_dict[activation] = activation_model
        except KeyError:
            pass
        self._layers = []

        self._default_name = _default_name(predictor)
        super().__init__(
            gp_model=gp_model,
            input_vars=input_vars,
            output_vars=output_vars,
            **kwargs,
        )

    def __iter__(self):
        """Iterate over layers of neural network"""
        return self.layers.__iter__()

    @property
    def layers(self):
        """Access models for successive layers of the network"""
        return self._layers

    def _add_dense_layer(
        self,
        input_vars,
        layer_coefs,
        layer_intercept,
        activation,
        activation_vars=None,
        **kwargs,
    ):
        """Add a dense layer to gurobipy model.

        Parameters
        ----------

        input_vars : mvar_array_like
            Decision variables used as input for predictor in model.
        layer_coefs:
            Coefficient for each node in a layer
        layer_intercept:
            Intercept bias
        activation:
            Activation function
        activation_vars : None, optional
            Output variables
        """
        layer = DenseLayer(
            self.gp_model,
            activation_vars,
            input_vars,
            layer_coefs,
            layer_intercept,
            activation,
            **kwargs,
        )
        self._layers.append(layer)
        return layer

    def _add_activation_layer(
        self, input_vars, activation, activation_vars=None, **kwargs
    ):
        """Add an activation layer to gurobipy model.

        Parameters
        ----------

        input_vars : mvar_array_like
            Decision variables used as input for predictor in gurobipy model.
        activation:
            Activation function
        activation_vars : mvar_array_like, optional
            Output variables
        """
        layer = ActivationLayer(
            self.gp_model, activation_vars, input_vars, activation, **kwargs
        )
        self._layers.append(layer)
        return layer

    def print_stats(self, abbrev=False, file=None):
        """Print statistics about submodel created.

        Parameters
        ----------

        file : None, optional
            Text stream to which output should be redirected. By default sys.stdout.
        """
        if not abbrev:
            super().print_stats(abbrev, file)
            print(file=file)

            self._print_container_steps("Layer", self._layers, file=file)
        else:
            for layer in self:
                layer.print_stats(abbrev=True, file=file)
                print(file=file)
